from cassandra.cluster import Cluster
from cassandra.query import SimpleStatement
from cassandra.auth import PlainTextAuthProvider
import pandas as pd
import numpy as np
import threading
import time
import logging as lg
from logger import create_logger

lg.getLogger("cassandra").setLevel('ERROR')

logger = create_logger("cql.ScyllaDB")

class ScyllaDB:
    def __init__(self,hosts:list,port:int,table_name:str,user:str,password:str) -> None:
        self.start_time = time.time() # for completion the time show
        # get keyspace & table name
        if len(table_name.split("."))!=2: raise Exception("Table name must be! keyspace.table_name")
        # keyspace
        self.keyspace_name = table_name.split(".")[0]
        # table nme
        self.table_name = table_name.split(".")[1]

        logger.debug(f"keyspace >> {self.keyspace_name}")
        logger.debug(f"table    >> {self.table_name}") 
        # set authenticator
        auth = PlainTextAuthProvider(username=user,password=password)
        # set cluster
        self.cluster = Cluster(contact_points=hosts,port=port,auth_provider=auth)
        # set keyspace
        self.session = self.cluster.connect(self.keyspace_name)
        logger.debug(f"connected to %s keyspace..."%(self.keyspace_name))
        stmt = SimpleStatement("SELECT column_name,type FROM system_schema.columns WHERE keyspace_name=%(key)s AND table_name=%(tbl)s")
        # get table's all columns and data type
        table_schema = self.session.execute(stmt,{ "key":self.keyspace_name,"tbl":self.table_name})
        logger.debug("%s columns fetched..."%(str(len([ i.column_name for i in table_schema]))))
        # has cql to python data type
        self.cql_to_python_type = { 
            'NULL': None,
            'boolean': pd.BooleanDtype(),
            'float': pd.Float64Dtype(),
            'double': pd.Float64Dtype(),
            'int': pd.Int64Dtype(),
            'smallint': pd.Int16Dtype(), 
            'tinyint': pd.Int64Dtype(),
            'counter': pd.Int64Dtype(),
            'varint': pd.Int64Dtype(),
            'bigint': pd.Int64Dtype(),
            'decimal': pd.Float64Dtype(),
            'ascii':str,
            'varchar':str,
            'text': str,
            'blob': str,
            'date': str,
            'timestamp': str,
            'time': str,
            'list': list,
            'set': set,
            'map': dict,
            'timeuuid,': None,
            'uuid': None 
        }
        
        self.cql_columns = []
        self.cql_types = {}
        # create dictonary for data frame: column_name:python_datatype
        for row in table_schema:
            self.cql_types[row.column_name] = self.cql_to_python_type[row.type]
            self.cql_columns.append(row.column_name)
		
        if len(self.cql_columns)!=0:
            param_keys = ",".join(self.cql_columns) 
            param_values = ",".join([ '%('+k+')s' for k in self.cql_columns])
            self.insert__statement = "INSERT INTO "+self.table_name+" ("+param_keys+") VALUES ("+param_values+")"
        else:
            logger.debug("columns set using file.")

    def execute_insert(self,df:pd.DataFrame)->None:
        
        # handle null for numeric type default 0 add
        df[df.select_dtypes(include=np.number).columns] = df.select_dtypes(include=np.number).fillna(0).astype(int) 
        # fill boolean
        df[df.select_dtypes(include=bool).columns] = df.select_dtypes(include=bool).fillna(False)
        # handle null for object and string types
        df[df.select_dtypes(exclude=np.number).columns] = df.select_dtypes(exclude=np.number).fillna('') 
       
        # empty list
        futures = []
        # excute async & store future object
        for row in df.to_dict(orient='records'): # iterate each records
            futures.append(self.session.execute_async(self.insert__statement,row))

        # wait for complete async 
        for future in futures:
            future.result() # wait until insert queries
        
    def insert(self,file_name:str,chuncks=10000,workers=4)->None:
        
        # supported files
        file_ext = file_name.split(".")[-1] # get extension
        if file_ext=="csv": # if csv then read csv with chuncks
            data_frames = pd.read_csv(file_name,chunksize=chuncks)
        elif file_ext=="json":
            data_frames = pd.read_json(file_name,chunksize=chuncks,lines=True)
        elif file_ext in ["xls","xlsx"]:
            data_frames = pd.read_excel(file_name,chunksize=chuncks)
        else:
            raise Exception("%s is not supported!"%(file_name))

        # empty list number of threads
        threads = list(range(workers))
        while True:
            try:
                # start thread
                for i in range(workers):
                    # fetch next dataframe
                    df = next(data_frames)
                    df.columns = map(str.lower,df.columns)
                    if len(self.cql_columns)==0:
                        self.cql_columns= df.columns.to_list()
                        param_keys = ",".join(self.cql_columns) 
                        param_values = ",".join([ '%('+k+')s' for k in self.cql_columns])
                        self.insert__statement = "INSERT INTO "+self.table_name+" ("+param_keys+") VALUES ("+param_values+")"
                    threads[i] = threading.Thread(target=self.execute_insert,args=(df,))
                    threads[i].start()  
                
                # wait until complete task
                for i in range(workers):
                    threads[i].join()
            # on execption break
            except Exception as e:
                print(e)
                break
        duration_minutes = (time.time() - self.start_time) / 60
        logger.debug(f"Completed in : {duration_minutes:.2f} minutes")



import threading
from cassandra.cluster import Cluster
from cassandra.auth import PlainTextAuthProvider
from collections import namedtuple
import pandas as pd
import numpy as np
import time
import logging as lg
from logger import create_logger

lg.getLogger("cassandra").setLevel('ERROR')

logger = create_logger("thread.AscyllaDB")


class AScyllaDB(threading.Thread):
    def __init__(self, host_list: list, port: int, keyspace: str, table_name: str, username: str, password: str, file_name='data.csv', chunks=10000, workers=4) -> None:
        self.start_time = time.time()
        super().__init__()
        auth = PlainTextAuthProvider(username, password)
        cluster = Cluster(host_list,
                          port, auth_provider=auth)
        logger.debug(
            f"Connection started... >> {str(host_list)}:{port} | keyspace: {keyspace} | table: {table_name}.")
        session = cluster.connect(keyspace)
        logger.debug(f"{keyspace} keyspace is connected.")

        Conf = namedtuple(
            "Conf", ["cluster", "session", "keyspace", "table_name", "file_name", "chunks", "workers"])
        self.conf = Conf(cluster, session, keyspace,
                         table_name, file_name, chunks, workers)

    def init_insert(self):
        data_frames = pd.read_csv(
            self.conf.file_name, chunksize=self.conf.chunks, dtype=self.dict_table_types)
        logger.debug("File is successfully loaded.")
        threads = list(range(self.conf.workers))
        logger.debug(f"{self.conf.workers} workers are ready.")

        def execute_insert(df: pd.DataFrame) -> None:
            # make lower columns name for cql
            df.columns = map(str.lower, df.columns)
            # handle null for numeric type default 0 add
            df[df.select_dtypes(include=np.number).columns] = df.select_dtypes(
                include=np.number).fillna(0).astype(int)
            # fill boolean
            df[df.select_dtypes(include=bool).columns] = df.select_dtypes(
                include=bool).fillna(False)
            # handle null for object and string types
            df[df.select_dtypes(exclude=np.number).columns] = df.select_dtypes(
                exclude=np.number).fillna('')
            # empty list
            futures = []
            # excute async & store future object
            for row in df.to_dict(orient='records'):  # iterate each records
                futures.append(self.conf.session.execute_async(
                    self.insert__statement, row))

            # wait for complete async
            for future in futures:
                future.result()  # wait until insert queries

        flag = 1
        count = 0
        while True:
            try:
                # start thread
                for i in range(self.conf.workers):
                    # fetch next dataframe
                    df = next(data_frames)
                    count += len(df)
                    df.columns = map(str.lower, df.columns)
                    if len(self.list_columns) == 0 and flag == 1:
                        self.list_columns = df.columns.to_list()
                        logger.warn("Assign column from the file.")
                        param_keys = ",".join(self.list_columns)
                        param_values = ",".join(
                            ['%('+k+')s' for k in self.list_columns])
                        self.insert__statement = "INSERT INTO "+self.conf.table_name + \
                            " ("+param_keys+") VALUES ("+param_values+")"
                        logger.debug(f"{self.insert} statement is generated.")
                        flag = 0
                    threads[i] = threading.Thread(
                        target=execute_insert, args=(df,))
                    threads[i].start()
                # wait until complete task
                for i in range(self.conf.workers):
                    threads[i].join()
            # on execption break
            except Exception as e:
                print(e)
                break
        duration_minutes = (time.time() - self.start_time) / 60
        logger.debug(
            f"DONE! {count} Records Transferred In {duration_minutes:.2f} Minutes.")

    def set_columns(self):
        logger.debug(f"Start fetch columns from {self.conf.table_name}.")
        table_schema = self.conf.session.execute(
            "SELECT column_name,type FROM system_schema.columns WHERE keyspace_name=%s AND table_name=%s", [self.conf.keyspace, self.conf.table_name])
        # has cql to python data type
        self.cql_to_python_type = {
            'NULL': None,
            'boolean': pd.BooleanDtype(),
            'float': pd.Float64Dtype(),
            'double': pd.Float64Dtype(),
            'int': pd.Int64Dtype(),
            'smallint': pd.Int16Dtype(),
            'tinyint': pd.Int64Dtype(),
            'counter': pd.Int64Dtype(),
            'varint': pd.Int64Dtype(),
            'bigint': pd.Int64Dtype(),
            'decimal': pd.Float64Dtype(),
            'ascii': str,
            'varchar': str,
            'text': str,
            'blob': str,
            'date': str,
            'timestamp': str,
            'time': str,
            'list': list,
            'set': set,
            'map': dict,
            'timeuuid,': None,
            'uuid': None
        }

        self.list_columns = []
        self.dict_table_types = {}
        for row in table_schema:
            self.dict_table_types[row.column_name] = self.cql_to_python_type[row.type]
            self.list_columns.append(row.column_name)

        if len(self.list_columns) != 0:
            param_keys = ",".join(self.list_columns)
            param_values = ",".join(['%('+k+')s' for k in self.list_columns])
            self.insert__statement = "INSERT INTO "+self.conf.table_name + \
                " ("+param_keys+") VALUES ("+param_values+")"
            n = len(self.list_columns)
            logger.debug(
                f"{n} columns are fetched from {self.conf.table_name}.")
        else:
            logger.warning(
                f"No columns found from table {self.conf.table_name}.")

    def run(self):
        self.set_columns()
        self.init_insert()



import logging as lg
lg.getLogger("cassandra").setLevel('ERROR')

def create_logger(name:str):

    formatter = lg.Formatter(
        '%(asctime)s  %(levelname)s  %(message)s')
    console_handler = lg.StreamHandler()
    console_handler.setLevel(lg.DEBUG)
    console_handler.setFormatter(formatter)
    logger = lg.getLogger(name)
    logger.addHandler(console_handler)
    logger.setLevel(lg.DEBUG)
    logger.propagate = False
    return logger


from cassandra.cluster import Cluster
from cassandra.query import SimpleStatement
from cassandra.auth import PlainTextAuthProvider
import pandas as pd
import numpy as np
import threading
import time
import sys
import logging as lg
from logger import create_logger

logger = create_logger("bulk_insert")

def insert(hosts:list,port:int,table_name="",user="cassandra",password="cassandra",file_name="data.csv",workers=4,chuncks=4):
    
    if len(sys.argv)<=1: 
        if table_name=="": raise Exception("table_name is required!")
    else:
        table_name=sys.argv[1] 

    start_time = time.time() 
    # get keyspace & table name
    if len(table_name.split("."))!=2: raise Exception("Table name must be! keyspace.table_name")
    # keyspace
    keyspace_name = table_name.split(".")[0]
    # table nme
    table_name = table_name.split(".")[1]

    logger.debug(f"keyspace >> {keyspace_name}")
    logger.debug(f"table    >> {table_name}") 
    # set authenticator
    auth = PlainTextAuthProvider(username=user,password=password)
    # set cluster
    cluster = Cluster(contact_points=hosts,port=port,auth_provider=auth)
    # set keyspace
    session = cluster.connect(keyspace_name)
    logger.debug("Connected to %s keyspace..."%(keyspace_name))
    stmt = SimpleStatement("SELECT column_name,type FROM system_schema.columns WHERE keyspace_name=%(key)s AND table_name=%(tbl)s")
    cql_columns = []
    cql_types = {}
    # has cql to python data type
    cql_to_python_type = { 
        'NULL': None,
        'boolean': pd.BooleanDtype(),
        'float': pd.Float64Dtype(),
        'double': pd.Float64Dtype(),
        'int': pd.Int64Dtype(),
        'smallint': pd.Int16Dtype(), 
        'tinyint': pd.Int64Dtype(),
        'counter': pd.Int64Dtype(),
        'varint': pd.Int64Dtype(),
        'bigint': pd.Int64Dtype(),
        'decimal': pd.Float64Dtype(),
        'ascii':str,
        'varchar':str,
        'text': str,
        'blob': str,
        'date': str,
        'timestamp': str,
        'time': str,
        'list': list,
        'set': set,
        'map': dict,
        'timeuuid,': None,
        'uuid': None 
    }
    # get table's all columns and data type
    table_schema = session.execute_async(stmt,{ "key":keyspace_name,"tbl":table_name})
    for row in table_schema.result():
        cql_types[row.column_name] = cql_to_python_type[row.type]
        cql_columns.append(row.column_name)

    insert__statement = ""
    if len(cql_columns)>=1:
        param_keys = ",".join(cql_columns) 
        param_values = ",".join([ '%('+k+')s' for k in cql_columns])
        insert__statement = "INSERT INTO "+table_name+" ("+param_keys+") VALUES ("+param_values+")"
    else:
        logger.debug("No columns found in table! columns fetch from csv.")

    def execute_insert(df:pd.DataFrame,session)->None:
        # handle null for numeric type default 0 add
        df[df.select_dtypes(include=np.number).columns] = df.select_dtypes(include=np.number).fillna(0).astype(int) 
        # fill boolean
        df[df.select_dtypes(include=bool).columns] = df.select_dtypes(include=bool).fillna(False)
        # handle null for object and string types
        df[df.select_dtypes(exclude=np.number).columns] = df.select_dtypes(exclude=np.number).fillna('') 
        
        # empty list
        futures = []
        # excute async & store future object
        for row in df.to_dict(orient='records'): # iterate each records
            futures.append(session.execute_async(insert__statement,row))

        # wait for complete async 
        for future in futures:
            future.result() # wait until insert queries
            
    data_frames = pd.read_csv(file_name,chunksize=chuncks)
    threads = list(range(workers))
    count = 0
    while True:
        try:
            # start thread
            for i in range(workers):
                # fetch next dataframe
                df = next(data_frames)
                count+=len(df)
                df.columns = map(str.lower,df.columns)
                if len(cql_columns)==0:
                    cql_columns = df.columns.to_list()
                    param_keys = ",".join(cql_columns) 
                    param_values = ",".join([ '%('+k+')s' for k in cql_columns])
                    insert__statement = "INSERT INTO "+table_name+" ("+param_keys+") VALUES ("+param_values+")"
                threads[i] = threading.Thread(target=execute_insert,args=(df,session,))
                threads[i].start()
            
            # wait until complete task
            for i in range(workers):
                threads[i].join()
        # on execption break
        except Exception as e:
            print(e)
            break

    duration_minutes = (time.time() - start_time) / 60
    logger.debug(
            f"DONE! {count} Records Transferred In {duration_minutes:.2f} Minutes.")
